@@ -33,6 +33,7 @@
 #include "qemu-common.h"
 #include "qemu-timer.h"
 #include "monitor.h"
+#include "sysemu.h"
 
 #include <dirent.h>
 #include <sys/ioctl.h>
@@ -68,9 +69,9 @@ typedef int USBScanFunc(void *opaque, int bus_num, int addr, int class_id,
 //#define DEBUG
 
 #ifdef DEBUG
-#define dprintf printf
+#define DPRINTF printf
 #else
-#define dprintf(...)
+#define DPRINTF(...)
 #endif
 
 #define USBDBG_DEVOPENED "husb: opened %s/devices\n"
@@ -105,15 +106,15 @@ enum {
 
 /*
  * Control transfer state.
- * Note that 'buffer' _must_ follow 'req' field because 
+ * Note that 'buffer' _must_ follow 'req' field because
  * we need contigious buffer when we submit control URB.
- */ 
+ */
 struct ctrl_struct {
     uint16_t len;
     uint16_t offset;
     uint8_t  state;
     struct   usb_ctrlrequest req;
-    uint8_t  buffer[2048];
+    uint8_t  buffer[8192];
 };
 
 struct USBAutoFilter {
@@ -132,6 +133,7 @@ typedef struct USBHostDevice {
     int       configuration;
     int       ninterfaces;
     int       closing;
+    Notifier  exit;
 
     struct ctrl_struct ctrl;
     struct endp_data endp_table[MAX_ENDPOINTS];
@@ -170,10 +172,10 @@ static void set_halt(USBHostDevice *s, int ep)
     s->endp_table[ep - 1].halted = 1;
 }
 
-/* 
+/*
  * Async URB state.
  * We always allocate one isoc descriptor even for bulk transfers
- * to simplify allocation and casts. 
+ * to simplify allocation and casts.
  */
 typedef struct AsyncURB
 {
@@ -220,41 +222,43 @@ static void async_complete(void *opaque)
     AsyncURB *aurb;
 
     while (1) {
-    	USBPacket *p;
+        USBPacket *p;
 
-	int r = ioctl(s->fd, USBDEVFS_REAPURBNDELAY, &aurb);
+        int r = ioctl(s->fd, USBDEVFS_REAPURBNDELAY, &aurb);
         if (r < 0) {
-            if (errno == EAGAIN)
+            if (errno == EAGAIN) {
                 return;
-
+            }
             if (errno == ENODEV && !s->closing) {
-                printf("husb: device %d.%d disconnected\n", s->bus_num, s->addr);
+                printf("husb: device %d.%d disconnected\n",
+                       s->bus_num, s->addr);
                 usb_host_close(s);
                 usb_host_auto_check(NULL);
                 return;
             }
 
-            dprintf("husb: async. reap urb failed errno %d\n", errno);
+            DPRINTF("husb: async. reap urb failed errno %d\n", errno);
             return;
         }
 
         p = aurb->packet;
 
-	dprintf("husb: async completed. aurb %p status %d alen %d\n", 
+        DPRINTF("husb: async completed. aurb %p status %d alen %d\n",
                 aurb, aurb->urb.status, aurb->urb.actual_length);
 
-	if (p) {
+        if (p) {
             switch (aurb->urb.status) {
             case 0:
                 p->len = aurb->urb.actual_length;
-                if (aurb->urb.type == USBDEVFS_URB_TYPE_CONTROL)
+                if (aurb->urb.type == USBDEVFS_URB_TYPE_CONTROL) {
                     async_complete_ctrl(s, p);
+                }
                 break;
 
             case -EPIPE:
                 set_halt(s, p->devep);
-		p->len = USB_RET_STALL;
-		break;
+                p->len = USB_RET_STALL;
+                break;
 
             default:
                 p->len = USB_RET_NAK;
@@ -262,7 +266,7 @@ static void async_complete(void *opaque)
             }
 
             usb_packet_complete(p);
-	}
+        }
 
         async_free(aurb);
     }
@@ -273,37 +277,38 @@ static void async_cancel(USBPacket *unused, void *opaque)
     AsyncURB *aurb = opaque;
     USBHostDevice *s = aurb->hdev;
 
-    dprintf("husb: async cancel. aurb %p\n", aurb);
+    DPRINTF("husb: async cancel. aurb %p\n", aurb);
 
     /* Mark it as dead (see async_complete above) */
     aurb->packet = NULL;
 
     int r = ioctl(s->fd, USBDEVFS_DISCARDURB, aurb);
     if (r < 0) {
-        dprintf("husb: async. discard urb failed errno %d\n", errno);
+        DPRINTF("husb: async. discard urb failed errno %d\n", errno);
     }
 }
 
 static int usb_host_claim_interfaces(USBHostDevice *dev, int configuration)
 {
     int dev_descr_len, config_descr_len;
-    int interface, nb_interfaces, nb_configurations;
+    int interface, nb_interfaces;
     int ret, i;
 
     if (configuration == 0) /* address state - ignore */
         return 1;
 
-    dprintf("husb: claiming interfaces. config %d\n", configuration);
+    DPRINTF("husb: claiming interfaces. config %d\n", configuration);
 
     i = 0;
     dev_descr_len = dev->descr[0];
-    if (dev_descr_len > dev->descr_len)
+    if (dev_descr_len > dev->descr_len) {
         goto fail;
-    nb_configurations = dev->descr[17];
+    }
 
     i += dev_descr_len;
     while (i < dev->descr_len) {
-        dprintf("husb: i is %d, descr_len is %d, dl %d, dt %d\n", i, dev->descr_len,
+        DPRINTF("husb: i is %d, descr_len is %d, dl %d, dt %d\n",
+                i, dev->descr_len,
                dev->descr[i], dev->descr[i+1]);
 
         if (dev->descr[i+1] != USB_DT_CONFIG) {
@@ -312,7 +317,7 @@ static int usb_host_claim_interfaces(USBHostDevice *dev, int configuration)
         }
         config_descr_len = dev->descr[i];
 
-	printf("husb: config #%d need %d\n", dev->descr[i + 5], configuration); 
+        printf("husb: config #%d need %d\n", dev->descr[i + 5], configuration);
 
         if (configuration < 0 || configuration == dev->descr[i + 5]) {
             configuration = dev->descr[i + 5];
@@ -323,7 +328,8 @@ static int usb_host_claim_interfaces(USBHostDevice *dev, int configuration)
     }
 
     if (i >= dev->descr_len) {
-        fprintf(stderr, "husb: update iface failed. no matching configuration\n");
+        fprintf(stderr,
+                "husb: update iface failed. no matching configuration\n");
         goto fail;
     }
     nb_interfaces = dev->descr[i + 4];
@@ -370,7 +376,7 @@ static int usb_host_release_interfaces(USBHostDevice *s)
 {
     int ret, i;
 
-    dprintf("husb: releasing interfaces\n");
+    DPRINTF("husb: releasing interfaces\n");
 
     for (i = 0; i < s->ninterfaces; i++) {
         ret = ioctl(s->fd, USBDEVFS_RELEASEINTERFACE, &i);
@@ -387,7 +393,7 @@ static void usb_host_handle_reset(USBDevice *dev)
 {
     USBHostDevice *s = DO_UPCAST(USBHostDevice, dev, dev);
 
-    dprintf("husb: reset device %u.%u\n", s->bus_num, s->addr);
+    DPRINTF("husb: reset device %u.%u\n", s->bus_num, s->addr);
 
     ioctl(s->fd, USBDEVFS_RESET);
 
@@ -400,6 +406,7 @@ static void usb_host_handle_destroy(USBDevice *dev)
 
     usb_host_close(s);
     QTAILQ_REMOVE(&hostdevs, s, next);
+    qemu_remove_exit_notifier(&s->exit);
 }
 
 static int usb_linux_update_endp_table(USBHostDevice *s);
@@ -416,15 +423,16 @@ static int usb_host_handle_data(USBHostDevice *s, USBPacket *p)
 
     urb = &aurb->urb;
 
-    if (p->pid == USB_TOKEN_IN)
-    	urb->endpoint = p->devep | 0x80;
-    else
-    	urb->endpoint = p->devep;
+    if (p->pid == USB_TOKEN_IN) {
+        urb->endpoint = p->devep | 0x80;
+    } else {
+        urb->endpoint = p->devep;
+    }
 
     if (is_halted(s, p->devep)) {
-	ret = ioctl(s->fd, USBDEVFS_CLEAR_HALT, &urb->endpoint);
+        ret = ioctl(s->fd, USBDEVFS_CLEAR_HALT, &urb->endpoint);
         if (ret < 0) {
-            dprintf("husb: failed to clear halt. ep 0x%x errno %d\n", 
+            DPRINTF("husb: failed to clear halt. ep 0x%x errno %d\n",
                    urb->endpoint, errno);
             return USB_RET_NAK;
         }
@@ -449,10 +457,11 @@ static int usb_host_handle_data(USBHostDevice *s, USBPacket *p)
 
     ret = ioctl(s->fd, USBDEVFS_SUBMITURB, urb);
 
-    dprintf("husb: data submit. ep 0x%x len %u aurb %p\n", urb->endpoint, p->len, aurb);
+    DPRINTF("husb: data submit. ep 0x%x len %u aurb %p\n",
+            urb->endpoint, p->len, aurb);
 
     if (ret < 0) {
-        dprintf("husb: submit failed. errno %d\n", errno);
+        DPRINTF("husb: submit failed. errno %d\n", errno);
         async_free(aurb);
 
         switch(errno) {
@@ -470,15 +479,16 @@ static int usb_host_handle_data(USBHostDevice *s, USBPacket *p)
 
 static int ctrl_error(void)
 {
-    if (errno == ETIMEDOUT)
+    if (errno == ETIMEDOUT) {
         return USB_RET_NAK;
-    else 
+    } else {
         return USB_RET_STALL;
+    }
 }
 
 static int usb_host_set_address(USBHostDevice *s, int addr)
 {
-    dprintf("husb: ctrl set addr %u\n", addr);
+    DPRINTF("husb: ctrl set addr %u\n", addr);
     s->dev.addr = addr;
     return 0;
 }
@@ -488,12 +498,12 @@ static int usb_host_set_config(USBHostDevice *s, int config)
     usb_host_release_interfaces(s);
 
     int ret = ioctl(s->fd, USBDEVFS_SETCONFIGURATION, &config);
- 
-    dprintf("husb: ctrl set config %d ret %d errno %d\n", config, ret, errno);
-    
-    if (ret < 0)
+
+    DPRINTF("husb: ctrl set config %d ret %d errno %d\n", config, ret, errno);
+
+    if (ret < 0) {
         return ctrl_error();
- 
+    }
     usb_host_claim_interfaces(s, config);
     return 0;
 }
@@ -506,13 +516,13 @@ static int usb_host_set_interface(USBHostDevice *s, int iface, int alt)
     si.interface  = iface;
     si.altsetting = alt;
     ret = ioctl(s->fd, USBDEVFS_SETINTERFACE, &si);
-    
-    dprintf("husb: ctrl set iface %d altset %d ret %d errno %d\n", 
-    	iface, alt, ret, errno);
-    
-    if (ret < 0)
-        return ctrl_error();
 
+    DPRINTF("husb: ctrl set iface %d altset %d ret %d errno %d\n",
+            iface, alt, ret, errno);
+
+    if (ret < 0) {
+        return ctrl_error();
+    }
     usb_linux_update_endp_table(s);
     return 0;
 }
@@ -524,16 +534,16 @@ static int usb_host_handle_control(USBHostDevice *s, USBPacket *p)
     int ret, value, index;
     int buffer_len;
 
-    /* 
+    /*
      * Process certain standard device requests.
      * These are infrequent and are processed synchronously.
      */
     value = le16_to_cpu(s->ctrl.req.wValue);
     index = le16_to_cpu(s->ctrl.req.wIndex);
 
-    dprintf("husb: ctrl type 0x%x req 0x%x val 0x%x index %u len %u\n",
-        s->ctrl.req.bRequestType, s->ctrl.req.bRequest, value, index, 
-        s->ctrl.len);
+    DPRINTF("husb: ctrl type 0x%x req 0x%x val 0x%x index %u len %u\n",
+            s->ctrl.req.bRequestType, s->ctrl.req.bRequest, value, index,
+            s->ctrl.len);
 
     if (s->ctrl.req.bRequestType == 0) {
         switch (s->ctrl.req.bRequest) {
@@ -546,8 +556,9 @@ static int usb_host_handle_control(USBHostDevice *s, USBPacket *p)
     }
 
     if (s->ctrl.req.bRequestType == 1 &&
-                  s->ctrl.req.bRequest == USB_REQ_SET_INTERFACE)
+                  s->ctrl.req.bRequest == USB_REQ_SET_INTERFACE) {
         return usb_host_set_interface(s, index, value);
+    }
 
     /* The rest are asynchronous */
 
@@ -562,12 +573,12 @@ static int usb_host_handle_control(USBHostDevice *s, USBPacket *p)
     aurb->hdev   = s;
     aurb->packet = p;
 
-    /* 
+    /*
      * Setup ctrl transfer.
      *
      * s->ctrl is layed out such that data buffer immediately follows
      * 'req' struct which is exactly what usbdevfs expects.
-     */ 
+     */
     urb = &aurb->urb;
 
     urb->type     = USBDEVFS_URB_TYPE_CONTROL;
@@ -580,10 +591,10 @@ static int usb_host_handle_control(USBHostDevice *s, USBPacket *p)
 
     ret = ioctl(s->fd, USBDEVFS_SUBMITURB, urb);
 
-    dprintf("husb: submit ctrl. len %u aurb %p\n", urb->buffer_length, aurb);
+    DPRINTF("husb: submit ctrl. len %u aurb %p\n", urb->buffer_length, aurb);
 
     if (ret < 0) {
-        dprintf("husb: submit failed. errno %d\n", errno);
+        DPRINTF("husb: submit failed. errno %d\n", errno);
         async_free(aurb);
 
         switch(errno) {
@@ -604,9 +615,10 @@ static int do_token_setup(USBDevice *dev, USBPacket *p)
     USBHostDevice *s = (USBHostDevice *) dev;
     int ret = 0;
 
-    if (p->len != 8)
+    if (p->len != 8) {
         return USB_RET_STALL;
- 
+    }
+
     memcpy(&s->ctrl.req, p->data, 8);
     s->ctrl.len    = le16_to_cpu(s->ctrl.req.wLength);
     s->ctrl.offset = 0;
@@ -614,17 +626,20 @@ static int do_token_setup(USBDevice *dev, USBPacket *p)
 
     if (s->ctrl.req.bRequestType & USB_DIR_IN) {
         ret = usb_host_handle_control(s, p);
-        if (ret < 0)
+        if (ret < 0) {
             return ret;
+        }
 
-        if (ret < s->ctrl.len)
+        if (ret < s->ctrl.len) {
             s->ctrl.len = ret;
+        }
         s->ctrl.state = CTRL_STATE_DATA;
     } else {
-        if (s->ctrl.len == 0)
+        if (s->ctrl.len == 0) {
             s->ctrl.state = CTRL_STATE_ACK;
-        else
+        } else {
             s->ctrl.state = CTRL_STATE_DATA;
+        }
     }
 
     return ret;
@@ -635,16 +650,17 @@ static int do_token_in(USBDevice *dev, USBPacket *p)
     USBHostDevice *s = (USBHostDevice *) dev;
     int ret = 0;
 
-    if (p->devep != 0)
+    if (p->devep != 0) {
         return usb_host_handle_data(s, p);
+    }
 
     switch(s->ctrl.state) {
     case CTRL_STATE_ACK:
         if (!(s->ctrl.req.bRequestType & USB_DIR_IN)) {
             ret = usb_host_handle_control(s, p);
-            if (ret == USB_RET_ASYNC)
+            if (ret == USB_RET_ASYNC) {
                 return USB_RET_ASYNC;
-
+            }
             s->ctrl.state = CTRL_STATE_IDLE;
             return ret > 0 ? 0 : ret;
         }
@@ -654,12 +670,14 @@ static int do_token_in(USBDevice *dev, USBPacket *p)
     case CTRL_STATE_DATA:
         if (s->ctrl.req.bRequestType & USB_DIR_IN) {
             int len = s->ctrl.len - s->ctrl.offset;
-            if (len > p->len)
+            if (len > p->len) {
                 len = p->len;
+            }
             memcpy(p->data, s->ctrl.buffer + s->ctrl.offset, len);
             s->ctrl.offset += len;
-            if (s->ctrl.offset >= s->ctrl.len)
+            if (s->ctrl.offset >= s->ctrl.len) {
                 s->ctrl.state = CTRL_STATE_ACK;
+            }
             return len;
         }
 
@@ -675,8 +693,9 @@ static int do_token_out(USBDevice *dev, USBPacket *p)
 {
     USBHostDevice *s = (USBHostDevice *) dev;
 
-    if (p->devep != 0)
+    if (p->devep != 0) {
         return usb_host_handle_data(s, p);
+    }
 
     switch(s->ctrl.state) {
     case CTRL_STATE_ACK:
@@ -691,12 +710,14 @@ static int do_token_out(USBDevice *dev, USBPacket *p)
     case CTRL_STATE_DATA:
         if (!(s->ctrl.req.bRequestType & USB_DIR_IN)) {
             int len = s->ctrl.len - s->ctrl.offset;
-            if (len > p->len)
+            if (len > p->len) {
                 len = p->len;
+            }
             memcpy(s->ctrl.buffer + s->ctrl.offset, p->data, len);
             s->ctrl.offset += len;
-            if (s->ctrl.offset >= s->ctrl.len)
+            if (s->ctrl.offset >= s->ctrl.len) {
                 s->ctrl.state = CTRL_STATE_ACK;
+            }
             return len;
         }
 
@@ -734,8 +755,9 @@ static int usb_host_handle_packet(USBDevice *s, USBPacket *p)
     }
 
     /* Rest of the PIDs must match our address */
-    if (s->state < USB_STATE_DEFAULT || p->devaddr != s->addr)
+    if (s->state < USB_STATE_DEFAULT || p->devaddr != s->addr) {
         return USB_RET_NODEV;
+    }
 
     switch (p->pid) {
     case USB_TOKEN_SETUP:
@@ -746,7 +768,7 @@ static int usb_host_handle_packet(USBDevice *s, USBPacket *p)
 
     case USB_TOKEN_OUT:
         return do_token_out(s, p);
- 
+
     default:
         return USB_RET_STALL;
     }
@@ -775,8 +797,9 @@ static int usb_linux_update_endp_table(USBHostDevice *s)
     }
 
     /* in address state */
-    if (configuration == 0)
+    if (configuration == 0) {
         return 1;
+    }
 
     /* get the desired configuration, interface, and endpoint descriptors
      * from device description */
@@ -786,7 +809,7 @@ static int usb_linux_update_endp_table(USBHostDevice *s)
 
     if (descriptors[i + 1] != USB_DT_CONFIG ||
         descriptors[i + 5] != configuration) {
-        dprintf("invalid descriptor data - configuration\n");
+        DPRINTF("invalid descriptor data - configuration\n");
         return 1;
     }
     i += descriptors[i];
@@ -822,15 +845,17 @@ static int usb_linux_update_endp_table(USBHostDevice *s)
         }
 
         /* advance to the endpoints */
-        while (i < length && descriptors[i +1] != USB_DT_ENDPOINT)
+        while (i < length && descriptors[i +1] != USB_DT_ENDPOINT) {
             i += descriptors[i];
+        }
 
         if (i >= length)
             break;
 
         while (i < length) {
-            if (descriptors[i + 1] != USB_DT_ENDPOINT)
+            if (descriptors[i + 1] != USB_DT_ENDPOINT) {
                 break;
+            }
 
             devep = descriptors[i + 2];
             switch (descriptors[i + 3] & 0x3) {
@@ -847,7 +872,7 @@ static int usb_linux_update_endp_table(USBHostDevice *s)
                 type = USBDEVFS_URB_TYPE_INTERRUPT;
                 break;
             default:
-                dprintf("usb_host: malformed endpoint type\n");
+                DPRINTF("usb_host: malformed endpoint type\n");
                 type = USBDEVFS_URB_TYPE_BULK;
             }
             s->endp_table[(devep & 0xf) - 1].type = type;
@@ -866,9 +891,9 @@ static int usb_host_open(USBHostDevice *dev, int bus_num,
     struct usbdevfs_connectinfo ci;
     char buf[1024];
 
-    if (dev->fd != -1)
+    if (dev->fd != -1) {
         goto fail;
-
+    }
     printf("husb: open device %d.%d\n", bus_num, addr);
 
     if (!usb_host_device_path) {
@@ -882,7 +907,7 @@ static int usb_host_open(USBHostDevice *dev, int bus_num,
         perror(buf);
         goto fail;
     }
-    dprintf("husb: opened %s\n", buf);
+    DPRINTF("husb: opened %s\n", buf);
 
     dev->bus_num = bus_num;
     dev->addr = addr;
@@ -899,21 +924,23 @@ static int usb_host_open(USBHostDevice *dev, int bus_num,
     {
         int x;
         printf("=== begin dumping device descriptor data ===\n");
-        for (x = 0; x < dev->descr_len; x++)
+        for (x = 0; x < dev->descr_len; x++) {
             printf("%02x ", dev->descr[x]);
+        }
         printf("\n=== end dumping device descriptor data ===\n");
     }
 #endif
 
 
-    /* 
-     * Initial configuration is -1 which makes us claim first 
+    /*
+     * Initial configuration is -1 which makes us claim first
      * available config. We used to start with 1, which does not
-     * always work. I've seen devices where first config starts 
+     * always work. I've seen devices where first config starts
      * with 2.
      */
-    if (!usb_host_claim_interfaces(dev, -1))
+    if (!usb_host_claim_interfaces(dev, -1)) {
         goto fail;
+    }
 
     ret = ioctl(fd, USBDEVFS_CONNECTINFO, &ci);
     if (ret < 0) {
@@ -924,20 +951,23 @@ static int usb_host_open(USBHostDevice *dev, int bus_num,
     printf("husb: grabbed usb device %d.%d\n", bus_num, addr);
 
     ret = usb_linux_update_endp_table(dev);
-    if (ret)
+    if (ret) {
         goto fail;
+    }
 
-    if (ci.slow)
+    if (ci.slow) {
         dev->dev.speed = USB_SPEED_LOW;
-    else
+    } else {
         dev->dev.speed = USB_SPEED_HIGH;
+    }
 
-    if (!prod_name || prod_name[0] == '\0')
+    if (!prod_name || prod_name[0] == '\0') {
         snprintf(dev->dev.product_desc, sizeof(dev->dev.product_desc),
                  "host:%d.%d", bus_num, addr);
-    else
+    } else {
         pstrcpy(dev->dev.product_desc, sizeof(dev->dev.product_desc),
                 prod_name);
+    }
 
     /* USB devio uses 'write' flag to check for async completions */
     qemu_set_fd_handler(dev->fd, NULL, async_complete, dev);
@@ -947,26 +977,38 @@ static int usb_host_open(USBHostDevice *dev, int bus_num,
 
 fail:
     dev->fd = -1;
-    if (fd != -1)
+    if (fd != -1) {
         close(fd);
+    }
     return -1;
 }
 
 static int usb_host_close(USBHostDevice *dev)
 {
-    if (dev->fd == -1)
+    if (dev->fd == -1) {
         return -1;
+    }
 
     qemu_set_fd_handler(dev->fd, NULL, NULL, NULL);
     dev->closing = 1;
     async_complete(dev);
     dev->closing = 0;
     usb_device_detach(&dev->dev);
+    ioctl(dev->fd, USBDEVFS_RESET);
     close(dev->fd);
     dev->fd = -1;
     return 0;
 }
 
+static void usb_host_exit_notifier(struct Notifier* n)
+{
+    USBHostDevice *s = container_of(n, USBHostDevice, exit);
+
+    if (s->fd != -1) {
+        ioctl(s->fd, USBDEVFS_RESET);
+    }
+}
+
 static int usb_host_initfn(USBDevice *dev)
 {
     USBHostDevice *s = DO_UPCAST(USBHostDevice, dev, dev);
@@ -974,6 +1016,8 @@ static int usb_host_initfn(USBDevice *dev)
     dev->auto_attach = 0;
     s->fd = -1;
     QTAILQ_INSERT_TAIL(&hostdevs, s, next);
+    s->exit.notify = usb_host_exit_notifier;
+    qemu_add_exit_notifier(&s->exit);
     usb_host_auto_check(NULL);
     return 0;
 }
@@ -1007,15 +1051,14 @@ USBDevice *usb_host_device_open(const char *devname)
 {
     struct USBAutoFilter filter;
     USBDevice *dev;
-    USBHostDevice *s;
     char *p;
 
     dev = usb_create(NULL /* FIXME */, "usb-host");
-    s = DO_UPCAST(USBHostDevice, dev, dev);
 
     if (strstr(devname, "auto:")) {
-        if (parse_filter(devname, &filter) < 0)
+        if (parse_filter(devname, &filter) < 0) {
             goto fail;
+        }
     } else {
         if ((p = strchr(devname, '.'))) {
             filter.bus_num    = strtoul(devname, NULL, 0);
@@ -1036,7 +1079,7 @@ USBDevice *usb_host_device_open(const char *devname)
     qdev_prop_set_uint32(&dev->qdev, "hostaddr",  filter.addr);
     qdev_prop_set_uint32(&dev->qdev, "vendorid",  filter.vendor_id);
     qdev_prop_set_uint32(&dev->qdev, "productid", filter.product_id);
-    qdev_init(&dev->qdev);
+    qdev_init_nofail(&dev->qdev);
     return dev;
 
 fail:
@@ -1051,13 +1094,13 @@ int usb_host_device_close(const char *devname)
     int bus_num, addr;
     USBHostDevice *s;
 
-    if (strstr(devname, "auto:"))
+    if (strstr(devname, "auto:")) {
         return usb_host_auto_del(devname);
-
-    if (usb_host_find_device(&bus_num, &addr, product_name, sizeof(product_name),
-                             devname) < 0)
+    }
+    if (usb_host_find_device(&bus_num, &addr, product_name,
+                                    sizeof(product_name), devname) < 0) {
         return -1;
-
+    }
     s = hostdev_find(bus_num, addr);
     if (s) {
         usb_device_delete_addr(s->bus_num, s->dev.addr);
@@ -1075,15 +1118,18 @@ static int get_tag_value(char *buf, int buf_size,
     const char *p;
     char *q;
     p = strstr(str, tag);
-    if (!p)
+    if (!p) {
         return -1;
+    }
     p += strlen(tag);
-    while (qemu_isspace(*p))
+    while (qemu_isspace(*p)) {
         p++;
+    }
     q = buf;
     while (*p != '\0' && !strchr(stopchars, *p)) {
-        if ((q - buf) < (buf_size - 1))
+        if ((q - buf) < (buf_size - 1)) {
             *q++ = *p;
+        }
         p++;
     }
     *q = '\0';
@@ -1118,51 +1164,62 @@ static int usb_host_scan_dev(void *opaque, USBScanFunc *func)
     device_count = 0;
     bus_num = addr = speed = class_id = product_id = vendor_id = 0;
     for(;;) {
-        if (fgets(line, sizeof(line), f) == NULL)
+        if (fgets(line, sizeof(line), f) == NULL) {
             break;
-        if (strlen(line) > 0)
+        }
+        if (strlen(line) > 0) {
             line[strlen(line) - 1] = '\0';
+        }
         if (line[0] == 'T' && line[1] == ':') {
             if (device_count && (vendor_id || product_id)) {
                 /* New device.  Add the previously discovered device.  */
                 ret = func(opaque, bus_num, addr, class_id, vendor_id,
                            product_id, product_name, speed);
-                if (ret)
+                if (ret) {
                     goto the_end;
+                }
             }
-            if (get_tag_value(buf, sizeof(buf), line, "Bus=", " ") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Bus=", " ") < 0) {
                 goto fail;
+            }
             bus_num = atoi(buf);
-            if (get_tag_value(buf, sizeof(buf), line, "Dev#=", " ") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Dev#=", " ") < 0) {
                 goto fail;
+            }
             addr = atoi(buf);
-            if (get_tag_value(buf, sizeof(buf), line, "Spd=", " ") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Spd=", " ") < 0) {
                 goto fail;
-            if (!strcmp(buf, "480"))
+            }
+            if (!strcmp(buf, "480")) {
                 speed = USB_SPEED_HIGH;
-            else if (!strcmp(buf, "1.5"))
+            } else if (!strcmp(buf, "1.5")) {
                 speed = USB_SPEED_LOW;
-            else
+            } else {
                 speed = USB_SPEED_FULL;
+            }
             product_name[0] = '\0';
             class_id = 0xff;
             device_count++;
             product_id = 0;
             vendor_id = 0;
         } else if (line[0] == 'P' && line[1] == ':') {
-            if (get_tag_value(buf, sizeof(buf), line, "Vendor=", " ") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Vendor=", " ") < 0) {
                 goto fail;
+            }
             vendor_id = strtoul(buf, NULL, 16);
-            if (get_tag_value(buf, sizeof(buf), line, "ProdID=", " ") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "ProdID=", " ") < 0) {
                 goto fail;
+            }
             product_id = strtoul(buf, NULL, 16);
         } else if (line[0] == 'S' && line[1] == ':') {
-            if (get_tag_value(buf, sizeof(buf), line, "Product=", "") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Product=", "") < 0) {
                 goto fail;
+            }
             pstrcpy(product_name, sizeof(product_name), buf);
         } else if (line[0] == 'D' && line[1] == ':') {
-            if (get_tag_value(buf, sizeof(buf), line, "Cls=", " (") < 0)
+            if (get_tag_value(buf, sizeof(buf), line, "Cls=", " (") < 0) {
                 goto fail;
+            }
             class_id = strtoul(buf, NULL, 16);
         }
     fail: ;
@@ -1173,8 +1230,9 @@ static int usb_host_scan_dev(void *opaque, USBScanFunc *func)
                    product_id, product_name, speed);
     }
  the_end:
-    if (f)
+    if (f) {
         fclose(f);
+    }
     return ret;
 }
 
@@ -1188,11 +1246,9 @@ static int usb_host_scan_dev(void *opaque, USBScanFunc *func)
  *
  * @return 0 failed, 1 succeeded ('line' contains data)
  */
-static int usb_host_read_file(char *line, size_t line_size, const char *device_file, const char *device_name)
+static int usb_host_read_file(char *line, size_t line_size,
+                              const char *device_file, const char *device_name)
 {
-#if 0
-    Monitor *mon = cur_mon;
-#endif
     FILE *f;
     int ret = 0;
     char filename[PATH_MAX];
@@ -1201,14 +1257,8 @@ static int usb_host_read_file(char *line, size_t line_size, const char *device_f
              device_file);
     f = fopen(filename, "r");
     if (f) {
-        fgets(line, line_size, f);
+        ret = fgets(line, line_size, f) != NULL;
         fclose(f);
-        ret = 1;
-#if 0
-    } else {
-        if (mon)
-            monitor_printf(mon, "husb: could not open %s\n", filename);
-#endif
     }
 
     return ret;
@@ -1239,59 +1289,71 @@ static int usb_host_scan_sys(void *opaque, USBScanFunc *func)
     while ((de = readdir(dir))) {
         if (de->d_name[0] != '.' && !strchr(de->d_name, ':')) {
             char *tmpstr = de->d_name;
-            if (!strncmp(de->d_name, "usb", 3))
+            if (!strncmp(de->d_name, "usb", 3)) {
                 tmpstr += 3;
+            }
             bus_num = atoi(tmpstr);
 
-            if (!usb_host_read_file(line, sizeof(line), "devnum", de->d_name))
+            if (!usb_host_read_file(line, sizeof(line), "devnum", de->d_name)) {
                 goto the_end;
-            if (sscanf(line, "%d", &addr) != 1)
+            }
+            if (sscanf(line, "%d", &addr) != 1) {
                 goto the_end;
-
+            }
             if (!usb_host_read_file(line, sizeof(line), "bDeviceClass",
-                                    de->d_name))
+                                    de->d_name)) {
                 goto the_end;
-            if (sscanf(line, "%x", &class_id) != 1)
+            }
+            if (sscanf(line, "%x", &class_id) != 1) {
                 goto the_end;
+            }
 
-            if (!usb_host_read_file(line, sizeof(line), "idVendor", de->d_name))
+            if (!usb_host_read_file(line, sizeof(line), "idVendor",
+                                    de->d_name)) {
                 goto the_end;
-            if (sscanf(line, "%x", &vendor_id) != 1)
+            }
+            if (sscanf(line, "%x", &vendor_id) != 1) {
                 goto the_end;
-
+            }
             if (!usb_host_read_file(line, sizeof(line), "idProduct",
-                                    de->d_name))
+                                    de->d_name)) {
                 goto the_end;
-            if (sscanf(line, "%x", &product_id) != 1)
+            }
+            if (sscanf(line, "%x", &product_id) != 1) {
                 goto the_end;
-
+            }
             if (!usb_host_read_file(line, sizeof(line), "product",
                                     de->d_name)) {
                 *product_name = 0;
             } else {
-                if (strlen(line) > 0)
+                if (strlen(line) > 0) {
                     line[strlen(line) - 1] = '\0';
+                }
                 pstrcpy(product_name, sizeof(product_name), line);
             }
 
-            if (!usb_host_read_file(line, sizeof(line), "speed", de->d_name))
+            if (!usb_host_read_file(line, sizeof(line), "speed", de->d_name)) {
                 goto the_end;
-            if (!strcmp(line, "480\n"))
+            }
+            if (!strcmp(line, "480\n")) {
                 speed = USB_SPEED_HIGH;
-            else if (!strcmp(line, "1.5\n"))
+            } else if (!strcmp(line, "1.5\n")) {
                 speed = USB_SPEED_LOW;
-            else
+            } else {
                 speed = USB_SPEED_FULL;
+            }
 
             ret = func(opaque, bus_num, addr, class_id, vendor_id,
                        product_id, product_name, speed);
-            if (ret)
+            if (ret) {
                 goto the_end;
+            }
         }
     }
  the_end:
-    if (dir)
+    if (dir) {
         closedir(dir);
+    }
     return ret;
 }
 
@@ -1316,7 +1378,7 @@ static int usb_host_scan(void *opaque, USBScanFunc *func)
             strcpy(devpath, USBDEVBUS_PATH);
             usb_fs_type = USB_FS_SYS;
             closedir(dir);
-            dprintf(USBDBG_DEVOPENED, USBSYSBUS_PATH);
+            DPRINTF(USBDBG_DEVOPENED, USBSYSBUS_PATH);
             goto found_devices;
         }
         f = fopen(USBPROCBUS_PATH "/devices", "r");
@@ -1325,7 +1387,7 @@ static int usb_host_scan(void *opaque, USBScanFunc *func)
             strcpy(devpath, USBPROCBUS_PATH);
             usb_fs_type = USB_FS_PROC;
             fclose(f);
-            dprintf(USBDBG_DEVOPENED, USBPROCBUS_PATH);
+            DPRINTF(USBDBG_DEVOPENED, USBPROCBUS_PATH);
             goto found_devices;
         }
         /* try additional methods if an access method hasn't been found yet */
@@ -1335,22 +1397,24 @@ static int usb_host_scan(void *opaque, USBScanFunc *func)
             strcpy(devpath, USBDEVBUS_PATH);
             usb_fs_type = USB_FS_DEV;
             fclose(f);
-            dprintf(USBDBG_DEVOPENED, USBDEVBUS_PATH);
+            DPRINTF(USBDBG_DEVOPENED, USBDEVBUS_PATH);
             goto found_devices;
         }
     found_devices:
         if (!usb_fs_type) {
-            if (mon)
+            if (mon) {
                 monitor_printf(mon, "husb: unable to access USB devices\n");
+            }
             return -ENOENT;
         }
 
         /* the module setting (used later for opening devices) */
         usb_host_device_path = qemu_mallocz(strlen(devpath)+1);
         strcpy(usb_host_device_path, devpath);
-        if (mon)
+        if (mon) {
             monitor_printf(mon, "husb: using %s file-system with %s\n",
                            fs_type[usb_fs_type], usb_host_device_path);
+        }
     }
 
     switch (usb_fs_type) {
@@ -1384,27 +1448,29 @@ static int usb_host_auto_scan(void *opaque, int bus_num, int addr,
     QTAILQ_FOREACH(s, &hostdevs, next) {
         f = &s->match;
 
-	if (f->bus_num > 0 && f->bus_num != bus_num)
+        if (f->bus_num > 0 && f->bus_num != bus_num) {
             continue;
-
-	if (f->addr > 0 && f->addr != addr)
+        }
+        if (f->addr > 0 && f->addr != addr) {
             continue;
+        }
 
-	if (f->vendor_id > 0 && f->vendor_id != vendor_id)
+        if (f->vendor_id > 0 && f->vendor_id != vendor_id) {
             continue;
+        }
 
-	if (f->product_id > 0 && f->product_id != product_id)
+        if (f->product_id > 0 && f->product_id != product_id) {
             continue;
-
+        }
         /* We got a match */
 
         /* Already attached ? */
-        if (s->fd != -1)
+        if (s->fd != -1) {
             return 0;
+        }
+        DPRINTF("husb: auto open: bus_num %d addr %d\n", bus_num, addr);
 
-        dprintf("husb: auto open: bus_num %d addr %d\n", bus_num, addr);
-
-	usb_host_open(s, bus_num, addr, product_name);
+        usb_host_open(s, bus_num, addr, product_name);
     }
 
     return 0;
@@ -1418,21 +1484,24 @@ static void usb_host_auto_check(void *unused)
     usb_host_scan(NULL, usb_host_auto_scan);
 
     QTAILQ_FOREACH(s, &hostdevs, next) {
-        if (s->fd == -1)
+        if (s->fd == -1) {
             unconnected++;
+        }
     }
 
     if (unconnected == 0) {
         /* nothing to watch */
-        if (usb_auto_timer)
+        if (usb_auto_timer) {
             qemu_del_timer(usb_auto_timer);
+        }
         return;
     }
 
     if (!usb_auto_timer) {
         usb_auto_timer = qemu_new_timer(rt_clock, usb_host_auto_check, NULL);
-        if (!usb_auto_timer)
+        if (!usb_auto_timer) {
             return;
+        }
     }
     qemu_mod_timer(usb_auto_timer, qemu_get_clock(rt_clock) + 2000);
 }
@@ -1462,13 +1531,15 @@ static int parse_filter(const char *spec, struct USBAutoFilter *f)
     f->product_id = 0;
 
     for (i = BUS; i < DONE; i++) {
-    	p = strpbrk(p, ":.");
-    	if (!p) break;
+        p = strpbrk(p, ":.");
+        if (!p) {
+            break;
+        }
         p++;
- 
-    	if (*p == '*')
-            continue;
 
+        if (*p == '*') {
+            continue;
+        }
         switch(i) {
         case BUS: f->bus_num = strtol(p, NULL, 10);    break;
         case DEV: f->addr    = strtol(p, NULL, 10);    break;
@@ -1514,8 +1585,9 @@ static const char *usb_class_str(uint8_t class)
 {
     const struct usb_class_info *p;
     for(p = usb_class_info; p->class != -1; p++) {
-        if (p->class == class)
+        if (p->class == class) {
             break;
+        }
     }
     return p->class_name;
 }
@@ -1545,13 +1617,15 @@ static void usb_info_device(Monitor *mon, int bus_num, int addr, int class_id,
     monitor_printf(mon, "  Device %d.%d, speed %s Mb/s\n",
                 bus_num, addr, speed_str);
     class_str = usb_class_str(class_id);
-    if (class_str)
+    if (class_str) {
         monitor_printf(mon, "    %s:", class_str);
-    else
+    } else {
         monitor_printf(mon, "    Class %02x:", class_id);
+    }
     monitor_printf(mon, " USB device %04x:%04x", vendor_id, product_id);
-    if (product_name[0] != '\0')
+    if (product_name[0] != '\0') {
         monitor_printf(mon, ", %s", product_name);
+    }
     monitor_printf(mon, "\n");
 }
 
@@ -1570,18 +1644,20 @@ static int usb_host_info_device(void *opaque, int bus_num, int addr,
 
 static void dec2str(int val, char *str, size_t size)
 {
-    if (val == 0)
+    if (val == 0) {
         snprintf(str, size, "*");
-    else
-        snprintf(str, size, "%d", val); 
+    } else {
+        snprintf(str, size, "%d", val);
+    }
 }
 
 static void hex2str(int val, char *str, size_t size)
 {
-    if (val == 0)
+    if (val == 0) {
         snprintf(str, size, "*");
-    else
+    } else {
         snprintf(str, size, "%04x", val);
+    }
 }
 
 void usb_host_info(Monitor *mon)
@@ -1591,8 +1667,10 @@ void usb_host_info(Monitor *mon)
 
     usb_host_scan(mon, usb_host_info_device);
 
-    if (QTAILQ_EMPTY(&hostdevs))
+    if (QTAILQ_EMPTY(&hostdevs)) {
         return;
+    }
+
     monitor_printf(mon, "  Auto filters:\n");
     QTAILQ_FOREACH(s, &hostdevs, next) {
         char bus[10], addr[10], vid[10], pid[10];
